/***************************************************************************
   Multi UART console serial I/O.
 
   This driver provides either polled or interrupt driven
   serial I/O for both the MCF5206 internal UARTs and the
   external 68hc901 "terminal" port on the board.
 
   Author:
     David Fiddes, D.J.Fiddes@hw.ac.uk
     http://www.calm.hw.ac.uk/davidf/coldfire/
 
   COPYRIGHT (c) 1989-1998.
   On-Line Applications Research Corporation (OAR).
   Copyright assigned to U.S. Government, 1994.
 
   The license and distribution terms for this file may be
   found in the file LICENSE in this distribution or at
 
   http://www.OARcorp.com/rtems/license.html.
  
   $Id: console.c,v 1.11 1998/02/19 23:02:02 joel Exp $
 ***************************************************************************/

#include <termios.h>
#include <bsp.h>
#include <rtems/libio.h>

/***************************************************************************/
/**  console driver configuration                                         **/

static const int use_mc68hc901_uart_interrupt = 0;
static const int use_internal_uart_interrupt  = 1;



/***************************************************************************/
/**  68hc901 uart polled driver routines                                  **/

static MC68HC901 *mc68hc901 = (MC68HC901 *)0x30000000;


/***************************************************************************
   Function : mc68hc901Initialise

   Description : This function carries out any hardware inistialisation
   required to operate the UART on the 68hc901.
 ***************************************************************************/
static void
mc68hc901Initialise (void)
{

  /* Do something profound by way of hardware setup here */

}


/***************************************************************************
   Function : mc68hc901Read

   Description : This reads a character from the 68hc901. It returns to the
   caller without blocking if not character is waiting.
 ***************************************************************************/
static int
mc68hc901Read (int minor)
{
	if ((mc68hc901->RSR & MC68HC901_RSR_BF) == 0)
		return -1;
		
	return mc68hc901->UDR;
}


/***************************************************************************
   Function : mc68hc901Write

   Description : This writes out each character in the buffer to the 
   68hc901 waiting till each one is sucessfully transmitted.
 ***************************************************************************/
static int
mc68hc901Write (int minor, const char *buf, int len)
{
	static char txBuf;
	
	while (len--) {
		while ((mc68hc901->TSR & MC68HC901_TSR_BE) == 0)
			continue;

 		txBuf = *buf++;
		mc68hc901->UDR = txBuf;
		mc68hc901->TSR = MC68HC901_TSR_HL_HIGH | MC68HC901_TSR_TE;
	}
	return 0;
}



/***************************************************************************/
/**  Internal UART driver routines                                        **/

struct IntUartInfoStruct {
	void        *ttyp;
	volatile uartstruct* dp;
};

static struct IntUartInfoStruct IntUartInfo[2];


/***************************************************************************
   Function : IntUartSetBaud

   Description : This computes the baud-rate-generator configuration register 
   value and sets it.
 ***************************************************************************/
static void
IntUartSetBaud (int baud, volatile uartstruct* uart )
{
	int divisor;

	divisor = MCF5206_CLOCK_SPEED / (32 * baud);

	if (divisor < 2 ) {
		divisor = 2;
	}

	uart->write.ubg1 = (divisor >> 8) & 0xFF;
	uart->write.ubg2 = divisor & 0xFF;

}


/***************************************************************************
   Function : IntUartSetBaud

   Description : This provides the hardware-dependent portion of tcsetattr().
   value and sets it. At the moment this just sets the baud rate.

   Note: The highest baudrate is 115200 as this stays within 
   an error of +/- 5% at 25MHz processor clock
 ***************************************************************************/
static int
IntUartSetAttributes (int minor, const struct termios *t)
{
	int baud;

	switch (t->c_cflag & CBAUD) {
	default:		baud = -1;		break;
	case B50:		baud = 50;		break;
	case B75:		baud = 75;		break;
	case B110:		baud = 110;		break;
	case B134:		baud = 134;		break;
	case B150:		baud = 150;		break;
	case B200:		baud = 200;		break;
	case B300:		baud = 300;		break;
	case B600:		baud = 600;		break;
	case B1200:		baud = 1200;	break;
	case B1800:		baud = 1800;	break;
	case B2400:		baud = 2400;	break;
	case B4800:		baud = 4800;	break;
	case B9600:		baud = 9600;	break;
	case B19200:	baud = 19200;	break;
	case B38400:	baud = 38400;	break;
	case B57600:	baud = 57600;	break;
	case B115200:	baud = 115200;	break;
	}
	if (baud > 0)
		IntUartSetBaud( baud, IntUartInfo[minor - 1].dp );
	return 0;
}


/***************************************************************************
   Function : IntUartInterruptHandler

   Description : This is the interrupt handler for the internal uart. It
   determines which channel caused the interrupt before queueing any received
   chars and dequeueing chars waiting for transmission.
 ***************************************************************************/
static rtems_isr
IntUartInterruptHandler (rtems_vector_number v)
{
	char ch;
	unsigned int chan = (v - UART1_IRQ_VECTOR) & 1;
	volatile uartstruct* uart = IntUartInfo[chan].dp;

	/*
	 * Buffer received?
	 */
	if (uart->read.uisr & MCF5206_UART_UISR_RXRDY) {
		while ((uart->read.usr & MCF5206_UART_USR_RXRDY) != 0) {
			ch = uart->read.urbuf;
			rtems_termios_enqueue_raw_characters( IntUartInfo[chan].ttyp, &ch, 1 );
		}
	}

	/*
	 * Buffer transmitted?
	 */
	if (uart->read.uisr & MCF5206_UART_UISR_TXRDY) {
		/* disable further TX interrupts */
		uart->write.uimr = MCF5206_UART_UIMR_FFULL;
		rtems_termios_dequeue_characters( IntUartInfo[chan].ttyp, 1 ); /* only one char at a time */
	}
}


/***************************************************************************
   Function : IntUartInitialise

   Description : This initialises the internal uart hardware for both
   internal uarts. If the internal uart is to be interrupt driven then the
   interrupt vectors are hooked.
 ***************************************************************************/
static void
IntUartInitialise (void)
{
	unsigned int ChanCount;
	volatile uartstruct* uart;

	for ( ChanCount = 0; ChanCount < 2; ChanCount++ )
	{	
		if (ChanCount == 0)
		{
			uart = &sim->uart1;
			IntUartInfo[ChanCount].dp = &sim->uart1;
		}
		else
		{
			uart = &sim->uart2;
			IntUartInfo[ChanCount].dp = &sim->uart2;
		}
	
		IntUartInfo[ChanCount].ttyp = NULL;

		/* Reset the UART */
		uart->write.ucr = MCF5206_UART_UCR_RESET_RX;
		uart->write.ucr = MCF5206_UART_UCR_RESET_TX;
		uart->write.ucr = MCF5206_UART_UCR_RESET_MR;
			  
		/* UMR1 */
		uart->write.umr  = MCF5206_UART_UMR1_RXRTS   |
			               MCF5206_UART_UMR1_ERR     |
        		           MCF5206_UART_UMR1_PM_NONE |
					       MCF5206_UART_UMR1_BC_8;
		/* UMR2 */
		uart->write.umr  = MCF5206_UART_UMR2_CM_NORMAL  |
						   MCF5206_UART_UMR2_STOP_BITS_1;

		if (ChanCount == 0)
		{
			sim->icr12 = MCF5206_SIM_ICR_IL(UART1_IRQ_LEVEL) |
				         MCF5206_SIM_ICR_IP(UART1_IRQ_PRIORITY);
		}
		else
		{
			sim->icr13 = MCF5206_SIM_ICR_IL(UART2_IRQ_LEVEL) |
				         MCF5206_SIM_ICR_IP(UART2_IRQ_PRIORITY);
		}

		     
		uart->write.uivr = UART1_IRQ_VECTOR + ChanCount;
		uart->write.uacr = 0;
		uart->write.uimr = 0;

		IntUartSetBaud( 19200, uart ); /* default to 19200 baud */

		uart->write.ucsr = 0xDD; /* internal baud rate generator */
			 
		uart->write.ucr = MCF5206_UART_UCR_TX_ENABLED |
			              MCF5206_UART_UCR_RX_ENABLED;
		if (use_internal_uart_interrupt)
		{
			rtems_isr_entry old_handler;
			rtems_status_code status;

			status = rtems_interrupt_catch (IntUartInterruptHandler,
										UART1_IRQ_VECTOR + ChanCount,
										&old_handler);
			uart->write.uimr = MCF5206_UART_UIMR_FFULL; /* Enable RX interrupts on char ready */

		}

	} /* of ChanCount loop */


	/* Enable UART1 and UART2 interrupts */
	if (use_internal_uart_interrupt)
	{
		sim->imr = sim->imr & ( ~(MCF5206_SIM_IMR_UART1 | MCF5206_SIM_IMR_UART2) );
	}

} /* IntUartInitialise */


/***************************************************************************
   Function : IntUartInterruptWrite

   Description : This writes a single character to the appropriate uart 
   channel. This is either called during an interrupt or in the user's task
   to initiate a transmit sequence. Calling this routine enables Tx 
   interrupts.
 ***************************************************************************/
static int
IntUartInterruptWrite (int minor, const char *buf, int len)
{
	volatile uartstruct* uart = IntUartInfo[ minor - 1 ].dp;

	uart->write.utbuf = *buf;
	uart->write.uimr = MCF5206_UART_UIMR_FFULL |
					   MCF5206_UART_UIMR_TXRDY;

	return 0;
}


/***************************************************************************
   Function : IntUartPollRead

   Description : This reads a character from the internal uart. It returns 
   to the caller without blocking if not character is waiting.
 ***************************************************************************/
static int
IntUartPollRead (int minor)
{
	volatile uartstruct* uart = IntUartInfo[ minor - 1 ].dp;

	if ((uart->read.usr & MCF5206_UART_USR_RXRDY) == 0)
		return -1;
		
	return uart->read.urbuf;
}


/***************************************************************************
   Function : IntUartPollWrite

   Description : This writes out each character in the buffer to the 
   appropriate internal uart channel waiting till each one is sucessfully 
   transmitted.
 ***************************************************************************/
static int
IntUartPollWrite (int minor, const char *buf, int len)
{
	volatile uartstruct* uart = IntUartInfo[ minor - 1 ].dp;

	while (len--) {
		static char txBuf;
		while ( (uart->read.usr & MCF5206_UART_USR_TXRDY) == 0 )
			continue;
		txBuf = *buf++;
		uart->write.utbuf = txBuf;
	}
	return 0;
}




/***************************************************************************/
/**  Console Driver Entry points                                          **/


/***************************************************************************
   Function : console_reserve_resources

   Description : This reserves resources consumed by this driver. It passes
   the request on to the termios subsystem.
 ***************************************************************************/
void console_reserve_resources(
  rtems_configuration_table *configuration
)
{
	rtems_termios_reserve_resources (configuration, 1);
}


/***************************************************************************
   Function : console_initialize

   Description : This initialises termios, both sets of uart hardware before
   registering /dev/tty devices for each channel and the system /dev/console.
 ***************************************************************************/
rtems_device_driver console_initialize(
  rtems_device_major_number  major,
  rtems_device_minor_number  minor,
  void                      *arg
)
{
	rtems_status_code status;

	/* Set up TERMIOS */
	rtems_termios_initialize ();

	/* Do device-specific initialization */
	mc68hc901Initialise ();

	IntUartInitialise (); 

	
	/* Register each device with a unique name */
	status = rtems_io_register_name ("/dev/tty00", major, 0);
	if (status != RTEMS_SUCCESSFUL)
		rtems_fatal_error_occurred (status);

	status = rtems_io_register_name ("/dev/tty01", major, 1);
	if (status != RTEMS_SUCCESSFUL)
		rtems_fatal_error_occurred (status);

	status = rtems_io_register_name ("/dev/tty02", major, 2);
	if (status != RTEMS_SUCCESSFUL)
		rtems_fatal_error_occurred (status);


	/* Register the console port */
	status = rtems_io_register_name ("/dev/console", major, CONSOLE_PORT);
	if (status != RTEMS_SUCCESSFUL)
		rtems_fatal_error_occurred (status);


	return RTEMS_SUCCESSFUL;
}


/***************************************************************************
   Function : console_open

   Description : This actually opens the device depending on the minor 
   number set during initialisation. The device specific access routines are
   passed to termios when the devices is opened depending on whether it is
   polled or not.
 ***************************************************************************/
rtems_device_driver console_open(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{
	rtems_status_code status;
	static const rtems_termios_callbacks mc68hc901Callbacks = {
		NULL,					/* firstOpen */
		NULL,					/* lastClose */
		mc68hc901Read,			/* pollRead */
		mc68hc901Write,			/* write */
		NULL,					/* setAttributes */
		NULL,					/* stopRemoteTx */
		NULL,					/* startRemoteTx */
		FALSE					/* outputUsesInterrupts */
	};

    static const rtems_termios_callbacks IntUartIntrCallbacks = {
		NULL,					/* firstOpen */
		NULL,					/* lastClose */
		NULL,					/* pollRead */
		IntUartInterruptWrite,	/* write */
		IntUartSetAttributes,	/* setAttributes */
		NULL,					/* stopRemoteTx */
		NULL,					/* startRemoteTx */
		TRUE					/* outputUsesInterrupts */
	};

	static const rtems_termios_callbacks IntUartPollCallbacks = {
		NULL,					/* firstOpen */
		NULL,					/* lastClose */
		IntUartPollRead,		/* pollRead */
		IntUartPollWrite,		/* write */
		IntUartSetAttributes,	/* setAttributes */
		NULL,					/* stopRemoteTx */
		NULL,					/* startRemoteTx */
		FALSE					/* outputUsesInterrupts */
	};


	/* open the port depending on the minor device number */
	switch ( minor )
	{
	
	case 0:
		status = rtems_termios_open (major, minor, arg, &mc68hc901Callbacks);
		break;

	case 1:
	case 2:
		if (use_internal_uart_interrupt)
		{
			rtems_libio_open_close_args_t *args = arg;
			status = rtems_termios_open (major, minor, arg, &IntUartIntrCallbacks);
			IntUartInfo[ minor - 1 ].ttyp = args->iop->data1;
		}
		else
		{
			status = rtems_termios_open (major, minor, arg, &IntUartPollCallbacks);
		}
		break;

	default:
		status = RTEMS_INVALID_NUMBER;
		break;

	}
	
	return status;
}
 

/***************************************************************************
   Function : console_close

   Description : This closes the device via termios
 ***************************************************************************/
rtems_device_driver console_close(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{
	return rtems_termios_close (arg);
}


/***************************************************************************
   Function : console_read

   Description : Read from the device via termios
 ***************************************************************************/
rtems_device_driver console_read(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{
	return rtems_termios_read (arg);
}


/***************************************************************************
   Function : console_write

   Description : Write to the device via termios
 ***************************************************************************/
rtems_device_driver console_write(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{
	return rtems_termios_write (arg);
}


/***************************************************************************
   Function : console_ioctl

   Description : Pass the IOCtl call to termios
 ***************************************************************************/
rtems_device_driver console_control(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void                    * arg
)
{ 
	return rtems_termios_ioctl (arg);
}
